package gdrivemount

import (
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"os"
	"strings"
	"time"

	"github.com/jacobsa/fuse/fuseops"
	drive "google.golang.org/api/drive/v3"
)

//FileEntry entry to handle files
type FileEntry struct {
	parent *FileEntry
	fs     *GDriveFS
	GFile  *drive.File
	isDir  bool
	Name   string // local name
	// fuseops
	Inode fuseops.InodeID
	Attr  fuseops.InodeAttributes

	// cache file
	tempFile *os.File
	// childs
	fileList []*FileEntry // children
}

func (fe *FileEntry) removeEntry(entry *FileEntry) {
	toremove := -1
	for i, v := range fe.fileList {
		if v == entry {
			toremove = i
			break
		}
	}

	if toremove == -1 {
		return
	}
	fe.fileList = append(fe.fileList[:toremove], fe.fileList[toremove+1:]...)
}

// useful for debug to count the entries
func (fe *FileEntry) count() int {
	count := 0

	for _, c := range fe.fileList {
		count += c.count()
	}
	return count + len(fe.fileList)

}

// Sync cached , upload
// Clear cache too
func (fe *FileEntry) sync() (err error) {
	if fe.tempFile == nil {
		return
	}
	fe.tempFile.Sync()
	fe.tempFile.Seek(0, io.SeekStart)

	ngFile := &drive.File{}

	up := fe.fs.srv.Files.Update(fe.GFile.Id, ngFile)
	_, err = up.Media(fe.tempFile).Do()

	return

}
func (fe *FileEntry) clearCache() (err error) {
	if fe.tempFile == nil {
		return
	}
	//fe.sync() // Sync?
	fe.tempFile.Close()
	os.Remove(fe.tempFile.Name())
	fe.tempFile = nil
	return
}
func (fe *FileEntry) cache() *os.File {
	if fe.tempFile != nil {
		return fe.tempFile
	}
	var res *http.Response
	var err error
	// Export GDocs (Special google doc documents needs to be exported make a config somewhere for this)
	switch fe.GFile.MimeType {
	case "application/vnd.google-apps.document":
		log.Println("Exporting as: text/markdown")
		res, err = fe.fs.srv.Files.Export(fe.GFile.Id, "text/plain").Download()
	case "application/vnd.google-apps.spreadsheet":
		log.Println("Exporting as: text/csv")
		res, err = fe.fs.srv.Files.Export(fe.GFile.Id, "text/csv").Download()
	default:
		res, err = fe.fs.srv.Files.Get(fe.GFile.Id).Download()
	}

	if err != nil {
		log.Println("MimeType:", fe.GFile.MimeType)
		log.Println("Error from GDrive API", err)
		return nil
	}
	defer res.Body.Close()

	// Local copy
	fe.tempFile, err = ioutil.TempFile(os.TempDir(), "gdfs")
	if err != nil {
		log.Println("Error creating temp file")
		return nil
	}
	io.Copy(fe.tempFile, res.Body)

	fe.tempFile.Seek(0, io.SeekStart)
	return fe.tempFile

}

// Load append whatever?
// append file to this tree
func (fe *FileEntry) appendGFile(f *drive.File) *FileEntry {

	name := f.Name
	count := 1
	nameParts := strings.Split(f.Name, ".")
	for {
		en := fe.findByName(name, false) // locally only
		if en == nil {                   // ok we want no value
			break
		}
		count++
		if len(nameParts) > 1 {
			name = fmt.Sprintf("%s(%d).%s", nameParts[0], count, strings.Join(nameParts[1:], "."))
		} else {
			name = fmt.Sprintf("%s(%d)", nameParts[0], count)
		}
	}

	// Create Attribute
	attr := fuseops.InodeAttributes{}
	attr.Nlink = 1
	attr.Size = uint64(f.Size)
	// Temp
	attr.Uid = fe.fs.getUID()
	attr.Gid = fe.fs.getGID()
	attr.Crtime, _ = time.Parse(time.RFC3339, f.CreatedTime)
	attr.Mtime, _ = time.Parse(time.RFC3339, f.ModifiedTime)
	attr.Mode = os.FileMode(0644) // default

	if f.MimeType == "application/vnd.google-apps.folder" {
		attr.Mode = os.FileMode(0755) | os.ModeDir
	}
	// Create an entry

	// lock from find inode to fileList append
	entry := &FileEntry{
		parent: fe,
		fs:     fe.fs,
		GFile:  f,
		Name:   name,                    // we dont need name
		Inode:  fe.fs.findUnusedInode(), // Lock somehow
		Attr:   attr,
		//fileMap: map[string]*FileEntry{},
		fileList: []*FileEntry{},
	}

	fe.fileList = append(fe.fileList, entry)
	//fe.fileMap[f.Name] = entry

	return entry
}
func (fe *FileEntry) findByInode(inode fuseops.InodeID, recurse bool) *FileEntry {
	if inode == fe.Inode {
		return fe // return self
	}
	// Recurse??
	for _, e := range fe.fileList {
		if e.Inode == inode {
			return e
		}
		if recurse {
			re := e.findByInode(inode, recurse)
			if re != nil {
				return re
			}
		}
	}
	// For each child we findByInode
	return nil
}

func (fe *FileEntry) findByName(name string, recurse bool) *FileEntry {
	// Recurse??
	for _, e := range fe.fileList {
		if e.Name == name {
			return e
		}
		if recurse {
			re := e.findByName(name, recurse)
			if re != nil {
				return re
			}
		}
	}
	// For each child we findByInode
	return nil
}

func (fe *FileEntry) findByGID(gdriveID string, recurse bool) *FileEntry {
	// Recurse??
	for _, e := range fe.fileList {
		if e.GFile.Id == gdriveID {
			return e
		}
		if recurse {
			re := e.findByGID(gdriveID, recurse)
			if re != nil {
				return re
			}
		}
	}
	// For each child we findByInode
	return nil
}
