// gdrivemount implements a google drive fuse driver
package gdrivemount

import (
	"io"
	"os"
	"os/user"
	"strconv"
	"syscall"
	"time"

	"dev.hexasoftware.com/hxs/prettylog"

	"golang.org/x/net/context"

	drive "google.golang.org/api/drive/v3"
	"google.golang.org/api/googleapi"

	"github.com/jacobsa/fuse"
	"github.com/jacobsa/fuse/fuseops"
	"github.com/jacobsa/fuse/fuseutil"
)

var (
	log = prettylog.New("gdrivemount")
)

type fileHandle struct {
	handleID fuseops.HandleID
	entry    *FileEntry
	//tempFile     *os.File
	uploadOnDone bool
	// Testing
	entries []fuseutil.Dirent
	buf     []byte
}

/*type DirEntry struct {
	file *FileEntry
}*/

// GDriveFS handler
type GDriveFS struct {
	fuseutil.NotImplementedFileSystem
	srv *drive.Service

	osuser *user.User
	root   *FileEntry // hiearchy reference

	fileHandles map[fuseops.HandleID]*fileHandle
}

////////////////////////////////////////////////////////
// TOOLS & HELPERS
////////////////////////////////////////////////////////

// appendFile solve parent before add,
func (fs *GDriveFS) appendFile(f *drive.File) *FileEntry {

	fil := fs.root.findByGID(f.Id, true)
	if fil != nil { // ignore existing ID
		return nil
	}

	var entry *FileEntry
	if len(f.Parents) == 0 {
		entry = fs.root.appendGFile(f) // = append(fs.root.fileList, entry)
	}
	if len(f.Parents) > 1 {
		log.Println("This one has more than 1 parent:", f.Parents)
	}
	for _, parent := range f.Parents { // hierarchy add
		parentEntry := fs.root.findByGID(parent, true)
		if parentEntry == nil {
			log.Fatalln("Non existent parent", parent)
		}

		// Here
		entry = parentEntry.appendGFile(f)
	}
	return entry
}

func (fs *GDriveFS) defaultAttributes() fuseops.InodeAttributes {
	uid, err := strconv.Atoi(fs.osuser.Uid)
	if err != nil {
		panic("Cannot convert user to int?")
	}
	gid, err := strconv.Atoi(fs.osuser.Gid)
	if err != nil {
		panic("Cannot convert gid to int?")
	}

	return fuseops.InodeAttributes{
		Nlink: 1,
		Mode:  0644, // default regular file
		Uid:   uint32(uid),
		Gid:   uint32(gid),
	}
}

func (fs *GDriveFS) findUnusedHandle() *fileHandle {
	// Lock here instead

	var handle fuseops.HandleID

	for handle = 1; handle < 99999; handle++ {
		_, ok := fs.fileHandles[handle]
		if !ok {
			break
		}
	}
	fh := &fileHandle{handleID: handle}
	fs.fileHandles[handle] = fh

	return fh
}

func (fs *GDriveFS) findUnusedInode() fuseops.InodeID {
	var inode fuseops.InodeID
	for inode = 2; inode < 99999; inode++ {
		f := fs.root.findByInode(inode, true)
		if f == nil {
			return inode
		}
	}
	if inode == 4 {
		log.Println("Inode is 4")

	}
	log.Println("0 Inode ODD")
	return 0
}

func NewGDriveFS() *GDriveFS {

	osuser, err := user.Current()
	if err != nil {
		log.Fatalf("Unable to fetch current user:", err)
	}
	fs := &GDriveFS{}
	fs.osuser = osuser
	fs.srv = GetDriveService()
	fs.root = &FileEntry{
		fs: fs,
		Attr: fuseops.InodeAttributes{
			Mode:  os.FileMode(0755) | os.ModeDir,
			Nlink: 1,
			Size:  4096,
			Uid:   fs.getUID(),
			Gid:   fs.getGID(),
		},
		GFile: nil,
		Inode: fuseops.RootInodeID,
		Name:  "",
		//fileMap: map[string]*FileEntry{},
		fileList: []*FileEntry{},
		isDir:    true,
	}
	fs.fileHandles = map[fuseops.HandleID]*fileHandle{}

	entry := fs.root.appendGFile(&drive.File{Id: "0", Name: "Loading..."})
	entry.Attr.Mode = os.FileMode(0)

	go fs.refresh() // async fetch

	return fs
}

// Cache somewhere
func (fs *GDriveFS) getUID() uint32 {
	uid, _ := strconv.Atoi(fs.osuser.Uid)
	return uint32(uid)
}
func (fs *GDriveFS) getGID() uint32 {
	gid, _ := strconv.Atoi(fs.osuser.Gid)
	return uint32(gid)
}

// Reload service files
func (fs *GDriveFS) refresh() {
	fileMap := map[string]*drive.File{} // Temporary map by google drive ID

	gdFields := googleapi.Field("nextPageToken, files(id,name,size,mimeType,parents,createdTime,modifiedTime)")
	log.Println("Loading file entries from gdrive")
	r, err := fs.srv.Files.List().PageSize(1000).
		SupportsTeamDrives(true).
		IncludeTeamDriveItems(true).
		Fields(gdFields).
		Do()
	if err != nil {
		panic(err)
	}
	log.Println("Loaded:", len(r.Files))

	for _, f := range r.Files {
		fileMap[f.Id] = f
	}

	// Rest of the page
	for r.NextPageToken != "" {
		log.Println("Loading next page")
		r, err = fs.srv.Files.List().PageToken(r.NextPageToken).Do()
		if err != nil {
			panic(err)
		}
		for _, f := range r.Files {
			fileMap[f.Id] = f
		}
	}
	log.Println("Total entries:", len(fileMap))

	if err != nil || r == nil {
		log.Fatal("Unable to retrieve files", err)
	}

	fs.root.fileList = []*FileEntry{} // clear fileList
	// Helper func to recurse
	// Everything loaded we add to our entries
	var appendParentOf func(f *drive.File)
	appendParentOf = func(f *drive.File) {
		for _, pID := range f.Parents {
			parentFile, ok := fileMap[pID]
			if !ok {
				parentFile, err = fs.srv.Files.Get(pID).Do()
				if err != nil {
					panic(err)
				}
				fileMap[parentFile.Id] = parentFile
			}
			appendParentOf(parentFile) // Recurse
			fs.appendFile(parentFile)
		}
	}

	for _, f := range fileMap {
		appendParentOf(f) // Check parent first
		fs.appendFile(f)
	}
	log.Println("Refresh done")

	log.Println("File count:", fs.root.count())
}

// OpenDir return nil error allows open dir
func (fs *GDriveFS) OpenDir(ctx context.Context, op *fuseops.OpenDirOp) (err error) {

	entry := fs.root.findByInode(op.Inode, true)
	if entry == nil {
		return fuse.ENOENT
	}

	fh := fs.findUnusedHandle()
	fh.entry = entry
	op.Handle = fh.handleID

	return // No error allow, dir open
}

// ReadDir lists files into readdirop
func (fs *GDriveFS) ReadDir(ctx context.Context, op *fuseops.ReadDirOp) (err error) {
	fh, ok := fs.fileHandles[op.Handle]
	if !ok {
		log.Fatal("Handle does not exists")
	}

	if op.Offset == 0 { // Rebuild/rewind dir list
		fh.entries = []fuseutil.Dirent{}

		//fh.buf = make([]byte, 1000000) // Temp bigbuf somehow
		//written := 0

		for i, v := range fh.entry.fileList {
			fusetype := fuseutil.DT_File
			if v.isDir {
				fusetype = fuseutil.DT_Directory
			}
			dirEnt := fuseutil.Dirent{
				Inode:  v.Inode,
				Name:   v.Name,
				Type:   fusetype,
				Offset: fuseops.DirOffset(i) + 1,
			}
			//	written += fuseutil.WriteDirent(fh.buf[written:], dirEnt)
			fh.entries = append(fh.entries, dirEnt)
		}
	}

	// local Buftest

	// Raw copy
	/*log.Println("Just copy data off:", op.Offset)
	for {
		n := copy(op.Dst[op.BytesRead:], fh.buf[op.Offset:])
		if n == 0 {

			break
		}
		log.Println("Written:", n)
		op.BytesRead += n
	}*/

	index := int(op.Offset)
	if index > len(fh.entries) {
		return fuse.EINVAL
	}
	if index > 0 {
		index++
	}
	count := 0
	for i := index; i < len(fh.entries); i++ {
		n := fuseutil.WriteDirent(op.Dst[op.BytesRead:], fh.entries[i])
		//log.Println("Written:", n)
		if n == 0 {
			break
		}
		op.BytesRead += n
		count++
	}
	//log.Printf("Written %d Dirent: '%s' TotalSent: %d", count, fh.entry.Name, op.BytesRead)

	/*log.Println("Sending from offset:", op.Offset)
	entries = entries[op.Offset:]

	log.Println("Total entries for this tree:", len(dir.entry.fileList))

	count := 0
	offCount := int(op.Offset)
	// Resume at the specified offset into the array.
	for i, v := range entries {
		fusetype := fuseutil.DT_File
		if v.isDir {
			fusetype = fuseutil.DT_Directory
		}
		dirEnt := fuseutil.Dirent{
			Inode:  v.Inode,
			Name:   v.Name,
			Type:   fusetype,
			Offset: fuseops.DirOffset(op.Offset + fuseops.DirOffset(i+1)),
		}
		//log.Println("Entry offset:", dirEnt.Offset)
		n := fuseutil.WriteDirent(op.Dst[op.BytesRead:], dirEnt)
		if n == 0 { // let go
			log.Println("Broke the Write")
			break
		}

		count++
		offCount++
		op.BytesRead += n
	}
	log.Println("BytesRead:", op.BytesRead, " count:", count, " offCount:", offCount, " last:", len(dir.entry.fileList)-count)*/
	return
}

// SetInodeAttributes Not sure what attributes gdrive support we just leave this blank for now
func (fs *GDriveFS) SetInodeAttributes(ctx context.Context, op *fuseops.SetInodeAttributesOp) (err error) {
	return
}

//GetInodeAttributes return attributes
func (fs *GDriveFS) GetInodeAttributes(ctx context.Context, op *fuseops.GetInodeAttributesOp) (err error) {

	f := fs.root.findByInode(op.Inode, true)
	if f == nil {
		return fuse.ENOENT
	}
	op.Attributes = f.Attr
	op.AttributesExpiration = time.Now().Add(time.Minute)

	return
}

// ReleaseDirHandle deletes file handle entry
func (fs *GDriveFS) ReleaseDirHandle(ctx context.Context, op *fuseops.ReleaseDirHandleOp) (err error) {
	delete(fs.fileHandles, op.Handle)
	return
}

// LookUpInode based on Parent and Name we return a self cached inode
func (fs *GDriveFS) LookUpInode(ctx context.Context, op *fuseops.LookUpInodeOp) (err error) {

	parentFile := fs.root.findByInode(op.Parent, true) // true means transverse all
	if parentFile == nil {
		return fuse.ENOENT
	}

	now := time.Now()
	// Transverse only local
	f := parentFile.findByName(op.Name, false)
	if f == nil {
		return fuse.ENOENT
	}

	op.Entry = fuseops.ChildInodeEntry{
		Attributes:           f.Attr,
		Child:                f.Inode,
		AttributesExpiration: now.Add(time.Minute),
		EntryExpiration:      now.Add(time.Minute),
	}
	return
}

// StatFS basically allows StatFS to run
/*func (fs *GDriveFS) StatFS(ctx context.Context, op *fuseops.StatFSOp) (err error) {
	return
}*/

// ForgetInode allows to forgetInode
func (fs *GDriveFS) ForgetInode(ctx context.Context, op *fuseops.ForgetInodeOp) (err error) {
	return
}

// GetXAttr special attributes
func (fs *GDriveFS) GetXAttr(ctx context.Context, op *fuseops.GetXattrOp) (err error) {
	return
}

//////////////////////////////////////////////////////////////////////////
// File OPS
//////////////////////////////////////////////////////////////////////////

// OpenFile creates a temporary handle to be handled on read or write
func (fs *GDriveFS) OpenFile(ctx context.Context, op *fuseops.OpenFileOp) (err error) {
	f := fs.root.findByInode(op.Inode, true) // might not exists

	// Generate new handle
	fh := fs.findUnusedHandle()
	fh.entry = f

	op.Handle = fh.handleID
	op.UseDirectIO = true

	return
}

// ReadFile  if the first time we download the google drive file into a local temporary file
func (fs *GDriveFS) ReadFile(ctx context.Context, op *fuseops.ReadFileOp) (err error) {
	lf := fs.fileHandles[op.Handle]

	localFile := lf.entry.cache()
	op.BytesRead, err = localFile.ReadAt(op.Dst, op.Offset)
	if err != nil {
		log.Println("Err reading file", err)
	}
	if err == io.EOF { // fuse does not expect a EOF
		err = nil
	}

	return
}

// CreateFile creates empty file in google Drive and returns its ID and attributes, only allows file creation on 'My Drive'
func (fs *GDriveFS) CreateFile(ctx context.Context, op *fuseops.CreateFileOp) (err error) {

	parentFile := fs.root.findByInode(op.Parent, true)
	if parentFile == nil {
		err = fuse.ENOENT
		return
	}
	// Only write on child folders
	if parentFile.Inode == fuseops.RootInodeID {
		return syscall.EPERM
	}

	// Generate ID
	//genId, err := fs.srv.Files.GenerateIds().Count(1).Do()
	//id := genId.Ids[0]
	parents := []string{parentFile.GFile.Id}
	newFile := &drive.File{
		Parents: parents,
		Name:    op.Name,
	}
	createdFile, err := fs.srv.Files.Create(newFile).Do()
	if err != nil {
		err = fuse.EINVAL
		return
	}

	entry := parentFile.appendGFile(createdFile) // Add new created file
	if entry == nil {
		err = fuse.EINVAL
		return
	}
	log.Println("File Created", entry)

	localFile := entry.cache()
	if localFile == nil {
		return fuse.EINVAL
	}
	// Associate a temp file to a new handle
	// Local copy
	// Lock
	fh := fs.findUnusedHandle()
	fh.entry = entry
	fh.uploadOnDone = true
	//
	op.Handle = fh.handleID
	op.Entry = fuseops.ChildInodeEntry{
		Attributes:           entry.Attr,
		Child:                entry.Inode,
		AttributesExpiration: time.Now().Add(time.Minute),
		EntryExpiration:      time.Now().Add(time.Minute),
	}
	op.Mode = entry.Attr.Mode

	return
}

// WriteFile as ReadFile it creates a temporary file on first read
// Maybe the ReadFile should be called here aswell to cache current contents since we are using writeAt
func (fs *GDriveFS) WriteFile(ctx context.Context, op *fuseops.WriteFileOp) (err error) {
	lf, ok := fs.fileHandles[op.Handle]
	if !ok {
		err = fuse.EIO
		return
	}

	localFile := lf.entry.cache()
	if localFile == nil {
		return fuse.EINVAL
	}

	_, err = localFile.WriteAt(op.Data, op.Offset)
	if err != nil {
		err = fuse.EIO
		return
	}
	lf.uploadOnDone = true

	return
}

// ReleaseFileHandle closes and deletes any temporary files, upload in case if changed locally
func (fs *GDriveFS) ReleaseFileHandle(ctx context.Context, op *fuseops.ReleaseFileHandleOp) (err error) {
	lf := fs.fileHandles[op.Handle]

	if lf.uploadOnDone {
		err = lf.entry.sync()
		if err != nil {
			return fuse.EINVAL
		}
	}
	lf.entry.clearCache()
	delete(fs.fileHandles, op.Handle)

	//go fs.refresh()
	return
}

// Unlink remove file and remove from local cache entry
func (fs *GDriveFS) Unlink(ctx context.Context, op *fuseops.UnlinkOp) (err error) {
	parentEntry := fs.root.findByInode(op.Parent, true)
	if parentEntry == nil {
		return fuse.ENOENT
	}
	if parentEntry.Inode == fuseops.RootInodeID {
		return syscall.EPERM
	}

	fileEntry := parentEntry.findByName(op.Name, false)
	if fileEntry == nil {
		return fuse.ENOATTR
	}
	err = fs.srv.Files.Delete(fileEntry.GFile.Id).Do()
	if err != nil {
		return fuse.EIO
	}

	parentEntry.removeEntry(fileEntry)

	return
}

// FlushFile just returns no error, maybe upload should be handled here
func (fs *GDriveFS) FlushFile(ctx context.Context, op *fuseops.FlushFileOp) (err error) {
	return
}

// MkDir creates a directory on a parent dir
func (fs *GDriveFS) MkDir(ctx context.Context, op *fuseops.MkDirOp) (err error) {

	parentFile := fs.root.findByInode(op.Parent, true)
	if parentFile == nil {
		return fuse.ENOENT
	}
	if parentFile.Inode == fuseops.RootInodeID {
		return syscall.EPERM
	}

	// Should check existent first too
	fi, err := fs.srv.Files.Create(&drive.File{
		Parents:  []string{parentFile.GFile.Id},
		MimeType: "application/vnd.google-apps.folder",
		Name:     op.Name,
	}).Do()
	if err != nil {
		return fuse.ENOATTR
	}
	entry := parentFile.appendGFile(fi)
	if entry == nil {
		return fuse.EINVAL
	}

	op.Entry = fuseops.ChildInodeEntry{
		Attributes:           entry.Attr,
		Child:                entry.Inode,
		AttributesExpiration: time.Now().Add(time.Minute),
		EntryExpiration:      time.Now().Add(time.Microsecond),
	}

	return
}

// RmDir fuse implementation
func (fs *GDriveFS) RmDir(ctx context.Context, op *fuseops.RmDirOp) (err error) {

	parentFile := fs.root.findByInode(op.Parent, true)
	if parentFile == nil {
		return fuse.ENOENT
	}
	if parentFile.Inode == fuseops.RootInodeID {
		return syscall.EPERM
	}

	theFile := parentFile.findByName(op.Name, false)

	err = fs.srv.Files.Delete(theFile.GFile.Id).Do()
	if err != nil {
		return fuse.ENOTEMPTY
	}

	parentFile.removeEntry(theFile)

	// Remove from entry somehow

	return
}

// Rename fuse implementation
func (fs *GDriveFS) Rename(ctx context.Context, op *fuseops.RenameOp) (err error) {
	oldParentFile := fs.root.findByInode(op.OldParent, true)
	if oldParentFile == nil {
		return fuse.ENOENT
	}
	newParentFile := fs.root.findByInode(op.NewParent, true)
	if newParentFile == nil {
		return fuse.ENOENT
	}

	if oldParentFile.Inode == fuseops.RootInodeID || newParentFile.Inode == fuseops.RootInodeID {
		return syscall.EPERM
	}

	oldFile := oldParentFile.findByName(op.OldName, false)

	ngFile := &drive.File{
		Name: op.NewName,
	}

	updateCall := fs.srv.Files.Update(oldFile.GFile.Id, ngFile)
	if oldParentFile != newParentFile {
		updateCall.RemoveParents(oldParentFile.GFile.Id)
		updateCall.AddParents(newParentFile.GFile.Id)
	}
	updatedFile, err := updateCall.Do()

	oldParentFile.removeEntry(oldFile)
	newParentFile.appendGFile(updatedFile)

	return

}
